@@ -4,8 +4,7 @@ import logging
 import math
 import asyncio
 import time
-import datetime
-from datetime import timezone
+from datetime import datetime, timezone, timedelta
 import jwt
 
 logger = logging.getLogger("core.task")
@@ -17,7 +16,7 @@ def _generate_storage_path(powerlink_key: str, task_key: str, app_name: str) ->
         reversed_timestamp = max_timestamp - original_timestamp  # Reverse timestamp
 
         # Ensure timestamp is interpreted in UTC
-        folder_timestamp = datetime.datetime.fromtimestamp(original_timestamp / 1000, tz=timezone.utc).strftime("%Y%m%d%H%M%S")
+        folder_timestamp = datetime.fromtimestamp(original_timestamp / 1000, tz=timezone.utc).strftime("%Y%m%d%H%M%S")
 
         # Construct the storage path
         storage_path = f"{powerlink_key}/{reversed_timestamp}/{task_key}/{app_name}/{folder_timestamp}"
@@ -59,13 +58,8 @@ class Core:
         self.RUNTIME_VARIABLES: dict[str, str | None] = {}
     
     def is_token_valid(self) -> bool:
-        """
-        Checks if the current JWT token is valid and not expired.
-
-        Returns:
-            bool: True if the token exists and is not expired, False otherwise.
-        """
         if not self.JWT_TOKEN:
+            logger.info("[CORE] JWT Token is missing, cannot validate.")
             return False
         try:
             payload = jwt.decode(self.JWT_TOKEN, options={"verify_signature": False})
@@ -73,8 +67,15 @@ class Core:
             if not exp:
                 return False
             expiry_time = datetime.fromtimestamp(exp, timezone.utc)
-            return expiry_time > datetime.now(timezone.utc)
-        except Exception:
+            now = datetime.now(timezone.utc)
+
+            # Add small buffer to avoid edge re-logins
+            if expiry_time <= now + timedelta(seconds=5):
+                logger.warning(f"[CORE] Token near or past expiry: exp={expiry_time}, now={now}")
+                return False
+            return True
+        except Exception as e:
+            logger.error(f"[CORE] Failed to parse JWT: {e}")
             return False
         
     @classmethod
@@ -93,10 +94,18 @@ class Core:
         jwt_token = os.environ.get("AGENTIX_CORE_JWT_TOKEN")
         password = os.environ.get("CORE_PASSWORD")
 
+        print("📜 Initializing Core from environment variables ... ")
         if not agent_name:
-            raise ValueError("Environment variable 'AGENT_NAME' is required but not set.")
+            print("❌ Environment variable 'AGENT_NAME' is not set.")
+        
         if not core_api:
-            raise ValueError("Environment variable 'CORE_API' is required but not set.")
+            print("❌ Environment variable 'CORE_API' is not set.")
+
+        if not jwt_token:
+            print("❌ Environment variable 'AGENTIX_CORE_JWT_TOKEN' is not set. Will need to login to Core.")
+        
+        if not password:
+            print("❌ Environment variable 'CORE_PASSWORD' is not set. Will need to login to Core.")
 
         core = cls(
             AGENT_NAME=agent_name,
@@ -104,21 +113,6 @@ class Core:
             JWT_TOKEN=jwt_token,
         )
         core._PASSWORD = password
-
-        if not core.is_token_valid():
-            print ("[CORE] Token Expired! Trying to relogin ...")
-            if core._PASSWORD:
-                asyncio.run(core.login(core._PASSWORD))
-                print("[CORE] Re-login successful, JWT token updated.")
-                os.environ["AGENTIX_CORE_JWT_TOKEN"] = core.JWT_TOKEN  # Optional re-cache
-            else:
-                raise RuntimeError("JWT expired and no password available to re-login")
-        else:
-            print("[CORE] JWT Token is valid, no need to re-login.")
-
-        print(f"[CORE] Using JWT Token ... Fetching runtime variables for agent: {core.AGENT_NAME}")
-        asyncio.run(core.fetch_runtime_variables())
-        print(f"[CORE] ✅ Ready")
         return core
         
     #================================================================================================
@@ -142,7 +136,6 @@ class Core:
     #================================================================================================
     # Fetch Runtime Variables from Core
     #================================================================================================
-    #TODO: Decorator of ensure Token here
     async def fetch_runtime_variables(self):
         """
         Fetches runtime variables from Core backend and stores them in self.RUNTIME_VARIABLES.
@@ -197,8 +190,22 @@ class Core:
 
         logger.info(f"🔑 [LOGIN] Authenticating to Core API with user: {self.AGENT_NAME}")
         
-        if not PASSWORD:
+        if self.is_token_valid():
+            logger.info("✅ [LOGIN] JWT Token is valid, no need to re-login.")
+            await self.fetch_runtime_variables()
+            return True
+        else:
+            logger.info("❌ [LOGIN] JWT Token is expired or missing, proceeding with login.")
+
+        if not PASSWORD and not self._PASSWORD:
             raise ValueError("[LOGIN] PASSWORD must be provided to login.")
+        
+        if not self.CORE_API:
+            raise ValueError("[LOGIN] CORE_API must be set to login to Core.")
+        
+        if not self.AGENT_NAME:
+            raise ValueError("[LOGIN] AGENT_NAME must be set to login to Core.")
+        
 
         logger.info("🔐 [LOGIN] Using username and password for logging to Core.")
 
@@ -207,7 +214,7 @@ class Core:
 
         auth_payload = {
             "identifier": self.AGENT_NAME,
-            "password": PASSWORD
+            "password": PASSWORD or self._PASSWORD
         }
 
         max_retries = 3
@@ -220,6 +227,7 @@ class Core:
                         if login_response.status == 200 and "jwt" in login_json:
                             self.JWT_TOKEN = login_json["jwt"]
                             self._PASSWORD = PASSWORD  # Store password for future use
+                            os.environ["AGENTIX_CORE_JWT_TOKEN"] = self.JWT_TOKEN
                             logger.info(f"✅ [LOGIN] Login succeeded on attempt {attempt}")                            
                             break
                         else:
@@ -352,7 +360,7 @@ class Core:
     #================================================================================================
     # Submit Task in Core
     #================================================================================================
-    async def submit_task(self, JOB_COOKIE: dict, task_key: str, output: list[dict] | None = None) -> bool:
+    async def submit_task(self, JOB_COOKIE: dict, task_key: str, output: list[dict] | None = None, metadata: str = None) -> bool:
         """
         Submits a task in Core by calling the `/submit` endpoint.
 
@@ -383,7 +391,7 @@ class Core:
             headers = {"Referer": f"{self.CORE_API}"}
             request_body = {
                 "data": {
-                    "metadata": f"Submitting Task from {self.AGENT_NAME}"
+                    "metadata": f"Submitting Task from {self.AGENT_NAME} {metadata}"
                 }
             }
             
@@ -415,7 +423,7 @@ class Core:
     #================================================================================================
     # Reject Task in Core
     #================================================================================================
-    async def reject_task(self, JOB_COOKIE: dict, task_key: str, rejection_reason: str = "No reason provided") -> bool:
+    async def reject_task(self, JOB_COOKIE: dict, task_key: str, rejection_reason: str = "No reason provided", metadata: str = None) -> bool:
         """
         Rejects a task in Core by calling the `/reject` endpoint.
 
@@ -447,7 +455,7 @@ class Core:
             headers = {"Referer": f"{self.CORE_API}"}
             request_body = {
                 "data": {
-                    "metadata": f"Rejecting Task from {self.AGENT_NAME}",
+                    "metadata": f"Rejecting Task from {self.AGENT_NAME} {metadata}",
                     "reason": rejection_reason
                 }
             }
@@ -765,13 +773,13 @@ class Core:
         if not app_name:
             raise ValueError("❌ [SAVE-FILE] app_name is required and cannot be empty.")
 
-        if not storage_path:
-            storage_path = _generate_storage_path(powerlink_key= powerlink_key, task_key=task_key, app_name=app_name)
-            logger.info(f"📁 No storage_path provided. Generated storage path: {storage_path}")
-        
-        logger.info(f"📁 Saving file to task data: {file_name} at {storage_path}")
-
         try:
+            if not storage_path:
+                storage_path = _generate_storage_path(powerlink_key= powerlink_key, task_key=task_key, app_name=app_name)
+                logger.info(f"📁 No storage_path provided. Generated storage path: {storage_path}")
+            
+            logger.info(f"📁 Saving file to task data: {file_name} at {storage_path}")
+
             # 1. Upload File
             upload_url = f"{self.CORE_API}/v1/tasks/{task_key}/uploadFile"
             headers = {"Referer": self.CORE_API}
@@ -827,4 +835,132 @@ class Core:
 
         except Exception as e:
             logger.error(f"❌ [SAVE-FILE] Exception: {e}")
-            raise
\ No newline at end of file
+            raise
+
+
+    #================================================================================================
+    # Add task action in Core
+    #================================================================================================
+    async def add_task_action(
+        self,
+        JOB_COOKIE: dict,
+        task_key: str,
+        name: str,
+        origin_identity: str,
+        origin_name: str,
+        destination_identity: str,
+        destination_name: str,
+        input_data: dict,
+        output_data: dict,
+        status: str = "SUCCESS",
+        description: str = ""
+    ) -> bool:
+        """
+        Adds a task action entry for a specific task in Core.
+
+        Args:
+            JOB_COOKIE (dict): Dictionary of cookies required for session authentication.
+            task_key (str): Unique identifier of the task.
+            name (str): Action name (e.g., 'get_user_location').
+            origin_identity (str): Origin identity of the action.
+            origin_name (str): Human-readable name of origin.
+            destination_identity (str): Destination identity.
+            destination_name (str): Human-readable name of destination.
+            input_data (dict): Dictionary to send as the action input.
+            output_data (dict): Dictionary to send as the action output.
+            status (str): Status of the action, default is "SUCCESS".
+            description (str): Optional description of the action.
+
+        Returns:
+            bool: True if action is added successfully.
+        """
+        if not JOB_COOKIE or not isinstance(JOB_COOKIE, dict):
+            logger.error("❌ [TASK-ACTION] JOB_COOKIE is required and must be a non-empty dictionary.")
+            raise ValueError("❌ [TASK-ACTION] JOB_COOKIE is required and must be a non-empty dictionary.")
+
+        if not task_key:
+            logger.error("❌ [TASK-ACTION] task_key is required and cannot be empty.")
+            raise ValueError("❌ [TASK-ACTION] task_key is required and cannot be empty.")
+
+        try:
+            logger.info(f">>> [TASK-ACTION] Add Task Action to Task: {task_key}")
+
+            url = f"{self.CORE_API}/v1/tasks/{task_key}/actions"
+            headers = {"Referer": f"{self.CORE_API}"}
+
+            request_body = {
+                "data": {
+                    "name": name,
+                    "description": description,
+                    "originIdentity": origin_identity,
+                    "originName": origin_name,
+                    "destinationIdentity": destination_identity,
+                    "destinationName": destination_name,
+                    "input": input_data,
+                    "status": status,
+                    "output": output_data
+                }
+            }
+
+            async with aiohttp.ClientSession(cookies=JOB_COOKIE) as session:
+                async with session.post(url, headers=headers, json=request_body, allow_redirects=False) as response:
+                    response_json = await response.json() if "application/json" in response.headers.get("Content-Type", "") else None
+
+                    if response.status not in [200, 201, 202]:
+                        logger.error(f"❌ [TASK-ACTION] Failed to add task action. HTTP Status: {response.status} - {response_json}")
+                        raise RuntimeError(f"❌ [TASK-ACTION] Failed to add task action. HTTP Status: {response.status} - {response_json}")
+
+                    logger.info("✅ [TASK-ACTION] Task action added successfully!")
+                    return True
+
+        except Exception as e:
+            logger.error(f"❌ [TASK-ACTION] Error adding task action: {e}")
+            raise RuntimeError(f"❌ [TASK-ACTION] Error adding task action: {e}")
+
+
+    # ================================================================================================
+    # Append task data in Core
+    # ================================================================================================
+    async def append_task_data(
+        self,
+        JOB_COOKIE: dict,
+        task_key: str,
+        metadata: str,
+        input_data: list | None = None,
+        output_data: list | None = None
+    ) -> bool:
+        if not JOB_COOKIE or not isinstance(JOB_COOKIE, dict):
+            logger.error("❌ [APPEND-TASK] JOB_COOKIE is required and must be a non-empty dictionary.")
+            raise ValueError("❌ [APPEND-TASK] JOB_COOKIE is required and must be a non-empty dictionary.")
+
+        if not task_key:
+            logger.error("❌ [APPEND-TASK] task_key is required and cannot be empty.")
+            raise ValueError("❌ [APPEND-TASK] task_key is required and cannot be empty.")
+
+        try:
+            logger.info(f">>> [APPEND-TASK] Appending Data to Task: {task_key}")
+            url = f"{self.CORE_API}/v1/tasks/{task_key}/append"
+            headers = {"Referer": f"{self.CORE_API}"}
+
+            data_block = {"metadata": metadata}
+            if input_data:
+                data_block["input"] = input_data
+            if output_data:
+                data_block["output"] = output_data
+
+            request_body = {"data": data_block}
+
+            async with aiohttp.ClientSession(cookies=JOB_COOKIE) as session:
+                async with session.post(url, headers=headers, json=request_body, allow_redirects=False) as response:
+                    response_json = await response.json() if "application/json" in response.headers.get("Content-Type", "") else None
+
+                    if response.status not in [200, 201, 202]:
+                        logger.error(f"❌ [APPEND-TASK] Failed to append task data. HTTP Status: {response.status} - {response_json}")
+                        raise RuntimeError(f"❌ [APPEND-TASK] Failed to append task data. HTTP Status: {response.status} - {response_json}")
+
+                    logger.info("✅ [APPEND-TASK] Task data appended successfully!")
+                    return True
+
+        except Exception as e:
+            logger.error(f"❌ [APPEND-TASK] Error calling Append Task API: {e}")
+            raise RuntimeError(f"❌ [APPEND-TASK] Error calling Append Task API: {e}")
\ No newline at end of file
